/*
 * LazyImporter jEdit Plugin (ClassPathManager.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.etheridge.lazyimporter;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;

import org.etheridge.lazyimporter.classpath.ClassPath;
import org.etheridge.lazyimporter.utility.FileUtilities;
import org.gjt.sp.jedit.jEdit;

/** 
 * This class manages the class path used in the LazyImporter plugin.
 */
public class ClassPathManager 
{
  // singleton instance
  private static ClassPathManager msInstance;

  // other class members
  private static JarFileFilter msJarFileFilter = new JarFileFilter();
  
  // the RT.jar file (lazily initialized when required).
  private File mRTJarFile;
  
  // the current version of the class path
  private ClassPath mClassPath;
  
  // class path refresh polling thread
  private ClassPathRefreshThread mPollingThread;
  
  // the default value of how long between each class path refresh
  public static final int DEFAULT_POLLING_INTERVAL = 60; // seconds
  public static final int MINIMUM_POLLING_INTERVAL = 5; // seconds
  
  // whether or not the class path manager has been intially loaded
  private static boolean msInitialLoadComplete = false;
  
  private ClassPathManager()
  {
    // initially create an empty class path
    mClassPath = new ClassPath();
    
    // create the polling thread for the first time
    mPollingThread = new ClassPathRefreshThread();
  }
  
  public static synchronized ClassPathManager getInstance()
  {
    if (msInstance == null) {
      msInstance = new ClassPathManager();
    }
    return msInstance;
  }
  
  /**
   * @return whether the ClassPath has been initially loaded
   */
  public static synchronized boolean hasLoaded()
  {
    return msInitialLoadComplete;
  }
  
  public void stopClassPathPolling()
  {
    if (mPollingThread != null) {
      mPollingThread.stopPollingThread();
    }
  }
  
  /**
   * Refreshes the class path. This is done in a worker thread, so a call to 
   * getCurrentClassPath() may not result in the refreshed copy of the class path
   * straight away.
   */
  public void refreshClassPath()
  {
    // stop the current polling thread
    mPollingThread.stopPollingThread();
    
    // create a new polling thread
    mPollingThread = new ClassPathRefreshThread();
  }
 
  public synchronized ClassPath getCurrentClassPath()
  {
    return mClassPath;
  }
  
  //
  // Utility Methods
  //
  
  /**
   * Attempts to get the RT.jar java File representation from the local file system
   */
  public File getRTJarFile()
  {
    if (mRTJarFile == null) {
      try {
        // base location to search from is the java.home
        File javaHomeLocation = new File(System.getProperty("java.home"));
        return FileUtilities.findFile(javaHomeLocation, "rt.jar", msJarFileFilter);
      } catch (IOException ioException) {
        // do nothing, this just means we couldn't find an RT.JAR
        // file, so it will return null.
      }
    }
    return mRTJarFile;
  }
    
  private static class JarFileFilter implements FileFilter
  {
    public boolean accept(File pathname)
    {
      return pathname.isDirectory() || pathname.getName().endsWith(".jar");
    }
  }

  //
  // Private helper methods
  //
  
  /**
   * Recreates the classpath based upon known LazyImporter jEdit properties
   */
  private void recreateClassPath()
  {
    ClassPath localClassPath = new ClassPath();

    // add the user defined classpath element(s) to the ClassPath
    localClassPath.addElements(jEdit.getProperty(LazyImporterProperties.USER_DEFINED_CLASSPATH_STRING, ""));

    // add the rt.jar to the end of the classpath
    String rtJarString = null;    
    if (jEdit.getBooleanProperty(LazyImporterProperties.RTJAR_AUTO_DETECTED, true)) {
      rtJarString = ClassPathManager.getInstance().getRTJarFile().getAbsolutePath();
    } else if (jEdit.getBooleanProperty(LazyImporterProperties.RTJAR_USER_SPECIFIED, false)) {
      rtJarString = jEdit.getProperty(LazyImporterProperties.RTJAR_USER_SPECIFIED_STRING, "");
    }
    
    // if the rt.jar file was set, then add it to the classpath
    if (rtJarString != null) {
      localClassPath.addElement(rtJarString);
    }
    
    synchronized(this) {
      mClassPath = localClassPath;
    }
    
    synchronized(ClassPathManager.class) {
      msInitialLoadComplete = true;
    }
  }
    
  //
  // Inner Classes
  //
  
  private class ClassPathRefreshThread extends Thread
  {
    private volatile boolean mStopRequested;
    
    public ClassPathRefreshThread()
    {
      // start the thread on construction
      start();
    }
    
    public void run()
    {
      // reset flag
      mStopRequested = false;
      
      // continue looping until interrupted.
      while (!mStopRequested) {
        
        // do work here
        recreateClassPath();
        
        // explicity garbage collect
        System.gc();
        
        try {
          Thread.sleep(jEdit.getIntegerProperty(LazyImporterProperties.CLASSPATH_POLLING_INTERVAL, DEFAULT_POLLING_INTERVAL) * 1000);
        } catch (InterruptedException ie) {
          Thread.currentThread().interrupt(); // reassert
        }
      }
      
    }
    
    public void stopPollingThread()
    {
      // set flag
      mStopRequested = true;
      
      // interrupt the thread
      interrupt();
    }
  }
  
}
