/*
 * LazyImporter jEdit Plugin (LazyImporterPlugin.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
package org.etheridge.lazyimporter;
 
import java.awt.Point;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import org.etheridge.lazyimporter.ClassPathManager;
import org.etheridge.lazyimporter.gui.ImportClassPathMessage;
import org.etheridge.lazyimporter.gui.ImportListWindow;
import org.etheridge.lazyimporter.gui.ImportSelectionListener;
import org.etheridge.lazyimporter.gui.InitialLoadingFrame;
import org.etheridge.lazyimporter.gui.MonitoredTask;
import org.etheridge.lazyimporter.importer.ClassImportStatement;
import org.etheridge.lazyimporter.importer.drawing.DefaultDrawingStrategy;
import org.etheridge.lazyimporter.importer.drawing.DrawingStrategy;
import org.etheridge.lazyimporter.importer.drawing.GroupingDrawingStrategy;
import org.etheridge.lazyimporter.importer.drawing.ManualGroupingDrawingStrategy;
import org.etheridge.lazyimporter.importer.ImportStatement;
import org.etheridge.lazyimporter.importer.JavaClass;
import org.etheridge.lazyimporter.importer.ordering.AlphabeticalOrderingStrategy;
import org.etheridge.lazyimporter.importer.ordering.ImplicitImportsAtTopOrderingStrategy;
import org.etheridge.lazyimporter.importer.ordering.JavaImportsAtBeginningOrderingStrategy;
import org.etheridge.lazyimporter.importer.ordering.JavaImportsAtBottomOrderingStrategy;
import org.etheridge.lazyimporter.importer.ordering.OrderingStrategy;
import org.etheridge.lazyimporter.importer.PackageImportStatement;
import org.etheridge.lazyimporter.LazyImporterProperties;
import org.etheridge.lazyimporter.options.ClassPathOptionsPane;
import org.etheridge.lazyimporter.options.LayoutOptionsPane;
import org.etheridge.lazyimporter.options.MiscellaneousOptionsPane;
import org.etheridge.lazyimporter.options.OrderingOptionsPane;
import org.etheridge.lazyimporter.options.QuickImportOptionsPane;

import org.gjt.sp.jedit.EBMessage;
import org.gjt.sp.jedit.EBPlugin;
import org.gjt.sp.jedit.gui.OptionsDialog;
import org.gjt.sp.jedit.GUIUtilities;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.Marker;
import org.gjt.sp.jedit.msg.ViewUpdate;
import org.gjt.sp.jedit.OptionGroup;
import org.gjt.sp.jedit.textarea.JEditTextArea;
import org.gjt.sp.jedit.View;

/**
 * @author Matt Etheridge
 */
public class LazyImporterPlugin extends EBPlugin
{
  // Single instance of ClassPath, this is reset whenever the properties
  // have changed.
  private static ClassPathManager msClassPathManager;  
  
  // maps from jEdit Views to the single import listener for that view.
  private static Map msImportListenerMap = new HashMap();
  
  // maps from jEdit Views to the single ImportListWindow instance for that
  // view.
  private static Map msImportListWindowMap = new HashMap();

  //
  // jEdit framework/callback methods
  //
    
  /**
   * Called on jEdit startup
   */
  public void start()
  {
    // if the plugin is loaded by "deferred" loading, we should get all views 
    // and add a listener for each.  jEdit 4.1 loaded everything at startup
    // so we used to be able to do this inital load when we got a ViewUpdate
    // message, however, jEdit4.2 cannot rely on this.   
    View[] views = jEdit.getViews();
    for(int i = 0; i < views.length; i++) {
      createListener(views[i]);
    }    
        
    // initial creation of the class path manager starts the polling thread
    // so it is a good time to do this now.
    msClassPathManager = ClassPathManager.getInstance();
  }

  /**
   *  Description of the Method
   *
   * @param  dialog  Description of the Parameter
   */
  public void createOptionPanes(OptionsDialog dialog)
  {
     OptionGroup optionGroup = new OptionGroup("LazyImporter");
     optionGroup.addOptionPane(new ClassPathOptionsPane());
     optionGroup.addOptionPane(new OrderingOptionsPane());
     optionGroup.addOptionPane(new LayoutOptionsPane());
     optionGroup.addOptionPane(new MiscellaneousOptionsPane());
     optionGroup.addOptionPane(new QuickImportOptionsPane());
     dialog.addOptionGroup(optionGroup);
  }

  /**
   *  Description of the Method
   *
   * @param  menuItems  Description of the Parameter
   */
  public void createMenuItems(Vector menuItems)
  {
    menuItems.addElement(GUIUtilities.loadMenu("lazyimporter.menu"));
  }

  /**
   *  Description of the Method
   *
   * @param  message  Description of the Parameter
   */
  public void handleMessage(EBMessage message)
  {
    if (message instanceof ViewUpdate) {
      ViewUpdate viewUpdateMessage = (ViewUpdate) message;

      if (viewUpdateMessage.getWhat() == ViewUpdate.CREATED) {
        // add a import list listener for this view
        View createdView = viewUpdateMessage.getView();
        createListener(createdView);
      } else if (viewUpdateMessage.getWhat() == ViewUpdate.CLOSED) {
        // remove import listener from import window
        View closedView = ((ViewUpdate) message).getView();
        ViewImportListener listenerToRemove = (ViewImportListener) msImportListenerMap.get(closedView);
        ImportListWindow window = (ImportListWindow) msImportListWindowMap.get(closedView);
        window.removeImportSelectionListener(listenerToRemove);
        
        // remove import listener and import window from maps
        msImportListenerMap.remove(closedView);
        msImportListWindowMap.remove(closedView);
      }
      
    }
  }
  
  //
  // Action Methods - these are "called" from the actions.xml configuration file.
  //
    
  /** 
   * Removes Unused Imports
   */
  public static void removeUnusedImports(final View view)
  {
    if (!isJavaFile(view)) {
      return;
    }

    MonitoredTask monitoredTask = new MonitoredTask(view.getTextArea(), 
      jEdit.getProperty("LazyImporter.MonitoredTask.RemovingUnusedImports.Title"), 5)
    {
      protected void doWork()
      {
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.WaitingForClassPathToLoad"));
        showInitialLoadingWindowIfRequired(view);
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.StoringCaretPosition"));
        int currentCursorPosition = view.getTextArea().getCaretPosition();

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.CreatingJavaClass"));
        JavaClass javaClass = new JavaClass(view.getTextArea().getText());

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.RemovingUnusedImports"));
        javaClass.removeUnusedImports();

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.RedrawingCurrentClass"));
        redrawCurrentTextArea(javaClass, view.getTextArea(), currentCursorPosition);
      }
    };
    monitoredTask.start();
  }  
  
  
  /** 
   * Resolves the wildcard imports of the current class.
   */
  public static void resolvePackageImports(final View view)
  {
    if (!isJavaFile(view)) {
      return;
    }
    
    MonitoredTask monitoredTask = new MonitoredTask(view.getTextArea(), 
      jEdit.getProperty("LazyImporter.MonitoredTask.ResolvingPackageImports.Title"), 5)
    {
      protected void doWork()
      {
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.WaitingForClassPathToLoad"));
        showInitialLoadingWindowIfRequired(view);
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.StoringCaretPosition"));
        int currentCursorPosition = view.getTextArea().getCaretPosition();

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.CreatingJavaClass"));
        JavaClass javaClass = new JavaClass(view.getTextArea().getText());

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.ResolvingWildcardImports"));
        javaClass.resolvePackageImports(msClassPathManager.getCurrentClassPath(), jEdit.getBooleanProperty(LazyImporterProperties.MISC_DISPLAY_CURRENT_PACKAGE_IMPORTS, false));

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.RedrawingCurrentClass"));
        redrawCurrentTextArea(javaClass, view.getTextArea(), currentCursorPosition);
      }
    };
    monitoredTask.start();
  }
  
  /**
   * Displays the class path element that is under the cursor (if any).
   */
  public static void findImportInClassPath(final View view)
  {
    if (!isJavaFile(view)) {
      return;
    }
    
    MonitoredTask monitoredTask = new MonitoredTask(view.getTextArea(), 
      jEdit.getProperty("LazyImporter.MonitoredTask.FindingImportInClassPath.Title"), 4)
    {
      protected void doWork()
      {
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.WaitingForClassPathToLoad"));
        showInitialLoadingWindowIfRequired(view);
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.CreatingJavaClass"));
        int currentCaretPosition = view.getTextArea().getCaretPosition();
        JavaClass javaClass = new JavaClass(view.getTextArea().getText());
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.FindingImportStatement"));
        final ImportStatement importStatementAtPosition = 
          javaClass.getImportStatementAtPosition(currentCaretPosition);
        if (importStatementAtPosition == null) {
          setStatusBarMessage(view, jEdit.getProperty("LazyImporter.ErrorMessages.NoImportUnderCursor"));
          return; // return and stop worker thread
        }
        if (importStatementAtPosition instanceof PackageImportStatement) {
          setStatusBarMessage(view, jEdit.getProperty("LazyImporter.ErrorMessages.OnlyFindClassImports"));
          return; // return and stop worker thread
        }
    
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.FindingClassInClasspath"));
        // given the import statement, ask the class path which package this
        // class path element.
        final List classPathElementsContainingClass = 
          msClassPathManager.getCurrentClassPath().findClass(((ClassImportStatement)importStatementAtPosition).getFullyQualifiedClassName());
    
        // show message in event thread
        SwingUtilities.invokeLater(new Runnable()
        {
          public void run()
          {
              Object objectMessage = ImportClassPathMessage.constructClassPathMessage
                ((ClassImportStatement)importStatementAtPosition, classPathElementsContainingClass);
              JOptionPane.showMessageDialog(view, objectMessage);
          }
        });
      }
    };
    monitoredTask.start();
  }
  
  
  /**
   * Attempt to resolve any missing imports for classes found in the currently
   * imported classes.
   */
  public static void resolveMissingImportsFromExistingImports(final View view)
  {
    if (!isJavaFile(view)) {
      return;
    }
    
    MonitoredTask monitoredTask = new MonitoredTask(view.getTextArea(), 
      jEdit.getProperty("LazyImporter.MonitoredTask.ResolvingMissingImports.Title"), 5)
    {
      protected void doWork()
      {
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.WaitingForClassPathToLoad"));
        showInitialLoadingWindowIfRequired(view);
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.StoringCaretPosition"));
        int currentCursorPosition = view.getTextArea().getCaretPosition();

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.CreatingJavaClass"));
        JavaClass javaClass = new JavaClass(view.getTextArea().getText());

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.ResolvingMissingImports"));
        resolveMissingImportsFromExistingImports(javaClass, view);

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.RedrawingCurrentClass"));
        redrawCurrentTextArea(javaClass, view.getTextArea(), currentCursorPosition);
      }
    };
    monitoredTask.start();
  }
  
  /**
   * Performs the quick import functions specified by the user via configuration.
   */
  public static void quickImport(final View view)
  {
    if (!isJavaFile(view)) {
      return;
    }
        
    MonitoredTask monitoredTask = new MonitoredTask(view.getTextArea(), 
      jEdit.getProperty("LazyImporter.MonitoredTask.QuickImport.Title"), 6)
    {
      protected void doWork()
      {
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.WaitingForClassPathToLoad"));
        showInitialLoadingWindowIfRequired(view);
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.CreatingJavaClass"));
        int currentCursorPosition = view.getTextArea().getCaretPosition();
        JavaClass javaClass = new JavaClass(view.getTextArea().getText());
    
        if (jEdit.getBooleanProperty(LazyImporterProperties.QUICK_IMPORT_RESOLVE_PACKAGE_IMPORTS, false)) {
          updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.ResolvingWildcardImports"));
          javaClass.resolvePackageImports(msClassPathManager.getCurrentClassPath(), jEdit.getBooleanProperty(LazyImporterProperties.MISC_DISPLAY_CURRENT_PACKAGE_IMPORTS, false));
        }
         
        if (jEdit.getBooleanProperty(LazyImporterProperties.QUICK_IMPORT_RESOLVE_MISSING_IMPORTS, false)) {
          updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.ResolvingMissingImports"));
          resolveMissingImportsFromExistingImports(javaClass, view);
        }

        if (jEdit.getBooleanProperty(LazyImporterProperties.QUICK_IMPORT_REMOVE_UNUSED_IMPORTS, false)) {
          updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.RemovingUnusedImports"));
          javaClass.removeUnusedImports();
        }
       
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.RedrawingCurrentClass"));
        redrawCurrentTextArea(javaClass, view.getTextArea(), currentCursorPosition);
      }
    };
    monitoredTask.start();
  }
  
  /**
   * Converts any imports into wildcard imports (.*)
   */
  public static void convertToWildcardImports(final View view)
  {
    if (!isJavaFile(view)) {
      return;
    }
    
    MonitoredTask monitoredTask = new MonitoredTask(view.getTextArea(), 
      jEdit.getProperty("LazyImporter.MonitoredTask.ConvertingToWildcardImports.Title"), 5)
    {
      protected void doWork()
      {
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.WaitingForClassPathToLoad"));
        showInitialLoadingWindowIfRequired(view);
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.StoringCaretPosition"));
        int currentCursorPosition = view.getTextArea().getCaretPosition();

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.CreatingJavaClass"));
        JavaClass javaClass = new JavaClass(view.getTextArea().getText());

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.ConvertingToWildcard"));
        javaClass.convertExplicitImportsToPackageImports();

        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.RedrawingCurrentClass"));
        redrawCurrentTextArea(javaClass, view.getTextArea(), currentCursorPosition);
      }
    };
    monitoredTask.start();
  }
  
  
  /**
   * Adds an import statement for the classname that is currently under the
   * cursor.
   */
  public static void importClassUnderCursor(final View view)
  {
    if (!isJavaFile(view)) {
      return;
    }
    
    // get the currently selected word - if there is currently a selection do not
    // try and select anymore - otherwise, select the current word that the cursor
    // falls on.
    final JEditTextArea textArea = view.getTextArea();
    if (textArea.getSelectedText() == null ||
        textArea.getSelectedText().length() <= 0) {
      textArea.selectWord();    
    }

    final String selectedText = textArea.getSelectedText();
    
    MonitoredTask monitoredTask = new MonitoredTask(view.getTextArea(), 
      jEdit.getProperty("LazyImporter.MonitoredTask.ImportClassUnderCursor.Title"), 3)
    {
      protected void doWork()
      {
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.WaitingForClassPathToLoad"));
        showInitialLoadingWindowIfRequired(view);
        
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.FindingClassInClassPath"));
        // see if there is any imports that match the selected word
        final Set fullyQualifiedClassMatches = 
          msClassPathManager.getCurrentClassPath().findUnqualifiedClass(selectedText);
    
        // if there are no matches, just return and do nothing
        if (fullyQualifiedClassMatches.size() == 0) {
          setStatusBarMessage(view, 
            jEdit.getProperty("LazyImporter.ErrorMessages.NoMatchesInClassPath.1") 
            + selectedText + " " + 
            jEdit.getProperty("LazyImporter.ErrorMessages.NoMatchesInClassPath.2"));
          return;
        } 
            
        // found some matches
        updateMessage(jEdit.getProperty("LazyImporter.MonitoredTask.CreatingPossibleImports"));
        // for every match we have, create an import statement
        List possibleImportStatements = new ArrayList();
        for (Iterator i = fullyQualifiedClassMatches.iterator(); i.hasNext();) {
          String fullyQualifiedClassName = (String) i.next();
          possibleImportStatements.add(ImportStatement.create(fullyQualifiedClassName));
        }
    
        // if there is only ONE matching class in teh classpath, import it 
        // without popping up a window (if the user chose this option)
        if (jEdit.getBooleanProperty(LazyImporterProperties.MISC_DO_NOT_POPUP_WINDOW_ON_SINGLE_MATCH, false) &&
            fullyQualifiedClassMatches.size() == 1) {
          int currentCursorPosition = view.getTextArea().getCaretPosition();
          JavaClass javaClass = new JavaClass(view.getTextArea().getText());
          javaClass.overwriteImportStatement((ImportStatement)possibleImportStatements.get(0));
          redrawCurrentTextArea(javaClass, view.getTextArea(), currentCursorPosition);
          return;
        }
        
        // get the ImportListWindow for this view, reset the ImportStatements it is
        // going to show, and then show the window in the correct place.
        getImportListWindow(view).setImportStatements
          ((ImportStatement[])possibleImportStatements.toArray(new ImportStatement[0]));
        final Point offsetPoint = textArea.offsetToXY(textArea.getCaretPosition());
        SwingUtilities.convertPointToScreen(offsetPoint, textArea);
        
        getImportListWindow(view).setLocation(offsetPoint);
        getImportListWindow(view).show();
        getImportListWindow(view).setVisible(true);
       
        SwingUtilities.invokeLater(new Runnable()
        {
          public void run()
          {
            getImportListWindow(view).toFront();
          }
        });
      }
    };
    monitoredTask.start();
  }
  
  
  //
  // Private Helper Methods
  //

  private static void resolveMissingImportsFromExistingImports(JavaClass javaClass, View view)
  {
    // get package-class map from the class path
    Map packageClassMap = new HashMap(msClassPathManager.getCurrentClassPath().getPackageClassMap());
    
    // list of import statements to add
    Set importStatementsToAdd = new HashSet();
    
    // for each import statement in the current class, determien if there are
    // any other classes in its package to import.
    List importStatements = javaClass.getImportStatements();
    for (Iterator i = importStatements.iterator(); i.hasNext();) {
      ImportStatement currentImportStatement = (ImportStatement) i.next();
      String currentPackage = currentImportStatement.getPackageName();
      
      // if this package is in the classpath, see if any of its classes are
      // being imported, and if so, add a new import statement to the JavaClass.
      if (packageClassMap.containsKey(currentPackage)) {
        List classesInThisPackage = new ArrayList((List) packageClassMap.get(currentPackage));
        for (Iterator j = classesInThisPackage.iterator(); j.hasNext();) {
          String currentClass = (String) j.next();
          if (javaClass.findClassUsage(currentClass)) {
            ImportStatement newImportStatement = ImportStatement.create(currentPackage + "." + currentClass);
            if (!importStatements.contains(newImportStatement)) {
              importStatementsToAdd.add(newImportStatement);
            }
          }
        }
      }
    }
    
    // if there were any import statements to add then add them to teh JavaClass
    for (Iterator i = importStatementsToAdd.iterator(); i.hasNext();) {
      javaClass.addImportStatement((ImportStatement)i.next());
    }
  }
  
  private static void showInitialLoadingWindowIfRequired(View view)
  {
    // if the ClassPathManager has not loaded yet, then show this window
    if (!ClassPathManager.hasLoaded()) {
      InitialLoadingFrame initialLoadingFrame = new InitialLoadingFrame(view);
      
      while (!ClassPathManager.hasLoaded()) {
        try { Thread.sleep(300); } catch (Exception e) { }
      }
      
      initialLoadingFrame.setVisible(false);
    }
  }
  
  /**
   * Determines whether the current file being shown in the specified view is
   * a java source file (*.java).
   *
   * @param view the view to check the current file in
   * @return whether or not the current file in the view is a java source file.
   */
  private static boolean isJavaFile(View view) 
  {
    if (!(view.getBuffer().getName().toLowerCase().endsWith(".java"))) {
      setStatusBarMessage(view, jEdit.getProperty("LazyImporter.ErrorMessages.InvalidJavaClass"));
      return false;
    }
    return true;
  }
  
  /**
   * Redraws the import statements in the specified text area, according to
   * the user-specified drawing strategies.
   */
  private static void redrawCurrentTextArea(final JavaClass javaClass, final JEditTextArea textArea, final int caretPosition)
  {
    // store current marker positions and shortcuts
    // NOTE: all of this should probably be done in the event thread as well!
    final Map markerMap = new HashMap();
    for (Iterator i = textArea.getBuffer().getMarkers().iterator(); i.hasNext();) {
      Marker currentMarker = (Marker) i.next();
      markerMap.put
        (new Integer(textArea.getLineOfOffset(currentMarker.getPosition())), 
         new Character(currentMarker.getShortcut()));
    }
    
    // text size before any change
    final int beforeLineCount = textArea.getLineCount();
    
    // do stuff to textarea
    javaClass.setOrderingStrategies(getUserSpecifiedOrderingStrategies());
    javaClass.setDrawingStrategy(getUserSpecifiedDrawingStrategy());
    javaClass.redrawImportStatements();

    // actually update the jEditText area
    // NOTE: this is done in the event thread to avoid any nasty stuff 
    //       happening! ;)    
    SwingUtilities.invokeLater(new Runnable()
    {
      public void run()
      {
        // set the text of the class on the jEditTextArea
        textArea.setText(javaClass.getClassText());
        
        // set the caret position if it is still valid
        if (caretPosition >= 0 && caretPosition <= textArea.getText().length()) {
          textArea.setCaretPosition(caretPosition);
        }
        
        // calculate different in line count, so markers can move accordingly
        int difference = textArea.getLineCount() - beforeLineCount;
        
        // remove any spuriously added markers
        textArea.getBuffer().removeAllMarkers();
        
        // add markers after we've modified the text
        for (Iterator i = markerMap.keySet().iterator(); i.hasNext();) {
          Integer currentLineNumber = (Integer) i.next();
          Character currentShortcut = (Character) markerMap.get(currentLineNumber);
          int lineToInsertMarker = currentLineNumber.intValue() + difference;
          
          // make sure the marker wasn't on one of the imports and has been deleted!
          if (lineToInsertMarker >= 0 ) {
            int positionToInsertMarker = textArea.getLineStartOffset(lineToInsertMarker);
            textArea.getBuffer().addMarker(currentShortcut.charValue(), positionToInsertMarker);
          }
        }
      }
    });
    
    
  }
  
  /**
   * @return the ordering stratagies based upon user-specified properties
   */ 
  private static OrderingStrategy[] getUserSpecifiedOrderingStrategies()
  {
    List orderingStrategies = new ArrayList();
    
    if (jEdit.getBooleanProperty(LazyImporterProperties.SORTING_SORT_ALPHABETICALLY, false)) {
      orderingStrategies.add(new AlphabeticalOrderingStrategy());
    }
    
    if (jEdit.getBooleanProperty(LazyImporterProperties.SORTING_CURRENT_PACKAGE_IMPORTS_AT_TOP, false)) {
      orderingStrategies.add(new ImplicitImportsAtTopOrderingStrategy());
    }
    
    if (jEdit.getBooleanProperty(LazyImporterProperties.SORTING_JAVA_IMPORTS_AT_EXTREMITIES, false)) {
      if (jEdit.getBooleanProperty(LazyImporterProperties.SORTING_JAVA_IMPORTS_AT_TOP, false)) {
        orderingStrategies.add(new JavaImportsAtBeginningOrderingStrategy());  
      } else if (jEdit.getBooleanProperty(LazyImporterProperties.SORTING_JAVA_IMPORTS_AT_BOTTOM, false)) {
        orderingStrategies.add(new JavaImportsAtBottomOrderingStrategy());
      }
    }
   
    return (OrderingStrategy[]) orderingStrategies.toArray(new OrderingStrategy[0]);
  }
 
  private static DrawingStrategy getUserSpecifiedDrawingStrategy()
  {
    if (jEdit.getBooleanProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS, false)) {
      int groupingLevel = jEdit.getIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_LEVEL, GroupingDrawingStrategy.DEFAULT_GROUP_LEVEL);
      int groupingSpaces = jEdit.getIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_SPACING, GroupingDrawingStrategy.DEFAULT_GROUP_SPACING);
      return new GroupingDrawingStrategy(groupingLevel, groupingSpaces);
    } else if (jEdit.getBooleanProperty(LazyImporterProperties.MANUAL_GROUP_IMPORTS, false)) {
      return new ManualGroupingDrawingStrategy(jEdit.getProperty(LazyImporterProperties.MANUAL_GROUP_IMPORTS_STRING));
    }
    return new DefaultDrawingStrategy();
  }
  
  private static ImportListWindow getImportListWindow(View view)
  {
    ImportListWindow window = (ImportListWindow) msImportListWindowMap.get(view);
    if (window == null) {
      window = new ImportListWindow();
      msImportListWindowMap.put(view, window);
    }
    return window;
  }

  /**
   * Sets the statusbar message in the event thread.
   */
  private static void setStatusBarMessage(final View view, final String message) 
  {
    SwingUtilities.invokeLater(new Runnable()
    {
      public void run()
      {
        view.getStatus().setMessage(message);
      }
    });
  }

  
  private static void createListener(View view)
  {
    ViewImportListener listener = new ViewImportListener(view);
    msImportListenerMap.put(view, listener);
    getImportListWindow(view).addImportSelectionListener(listener);
  }
    
  //
  // Inner Classes
  //
  
  /**
   * Listens for import selection from popup import window.
   */
  private static class ViewImportListener implements ImportSelectionListener
  {
    private View mView;
    
    public ViewImportListener(View view)
    {
      mView = view;
    }
    
    public void importSelected(ImportStatement statementSelected)
    {
      int currentCursorPosition = mView.getTextArea().getCaretPosition();
      JavaClass javaClass = new JavaClass(mView.getTextArea().getText());
      javaClass.overwriteImportStatement(statementSelected);
      redrawCurrentTextArea(javaClass, mView.getTextArea(), currentCursorPosition);
    }
  }
  
}
