/*
 * LazyImporter jEdit Plugin (ClassPath.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
package org.etheridge.lazyimporter.classpath;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;

/**
 * This class represents a classpath.  It contains ClassPath elements.
 */
public class ClassPath
{
  // the list of elements (ClassPathElements).
  private List mElements;
  
  // package -> (list of) classes map
  private Map mPackageClassMap;
  
  // class -> (list of) packages map
  private Map mClassPackageMap;
  
  public ClassPath()
  {
    mElements = new ArrayList();
    mPackageClassMap = new HashMap();
    mClassPackageMap = new HashMap();
  }
  
  /**
   * Adds a classpath element to the classpath.
   *
   * @param element the classpath element to add - should be a string
   * representing a SINGLE element in the classpath.  For example, "D:\crap.jar"
   * or "D:\source"
   */
  public void addElement(String element) throws IllegalArgumentException
  {
    // if the string ends in a *.jar or *.zip then it is a JarFileElement, so
    // create one of these and add it to the list of elements
    element = element.toLowerCase();
    ClassPathElement classPathElement = null;
    if (element.endsWith(".zip") || element.endsWith(".jar")) {
      classPathElement = new JarFileElement(element, mElements.size());
    } else { // must be a path entry
      classPathElement = new DirElement(element, mElements.size()); 
    }
    
    // add to list of class path elements
    if (!mElements.contains(classPathElement)) {
      mElements.add(classPathElement);
    }
    
    updatePackageClassMap(classPathElement);
    updateClassPackageMap(classPathElement);
  }
  
  /** 
   * Adds multiple class path elements in string form.  This should be a 
   * stringified version of the classpath, with each class path element
   * separated by a path separator (ie. ; on windows, : on unix).
   *
   * For example, "D:\progs\jEdit4.0\jEdit.jar;D:\classes"
   * 
   * @param elements the stringified classpath elements to remove.
   */
  public void addElements(String elements)
  {
    // tokenize on the path separator
    String pathSeparator = System.getProperty("path.separator");
    StringTokenizer tokenizer = new StringTokenizer(elements, pathSeparator);
    while (tokenizer.hasMoreTokens()) {
      addElement(tokenizer.nextToken());
    }
    
  }
  
  /**
   * Removes the specified classpath element from the class path if it 
   * exists.
   *
   * @param element the classpath element (string) to remove.
   * @return whether or not the element was removed.
   */
  public boolean removeElement(String element)
  {
    if (mElements != null) {
      for (Iterator i = mElements.iterator(); i.hasNext();) {
        ClassPathElement currentElement = (ClassPathElement) i.next();
        if (currentElement.getName().equalsIgnoreCase(element)) {
          i.remove();
          return true;
        }
      }
    }
    return false;
  }
  
  /**
   * Clears all of the class path elements in this classpath.
   */
  public void clearElements()
  {
    mElements.clear();
    mPackageClassMap.clear();
  }
  
  public String getClassPathString()
  {
    StringBuffer buffer = new StringBuffer();
    for (Iterator i = mElements.iterator(); i.hasNext();) {
      buffer.append(((ClassPathElement)i.next()).getName());
      buffer.append(System.getProperty("path.separator"));
    }
    return buffer.toString();
  }
  
  /**
   * Returns a map that contains mappings between java packages and their
   * containing classes that are found in this class path.
   * 
   * NOTE: returning an unmodifiableMap, means only the map is unmodifiable -
   *       any Lists or other data structures inside the map can still be
   *       modified. (bit dodgy).
   * 
   * @return package class mapping.
   */
  public Map getPackageClassMap()
  {
    return Collections.unmodifiableMap(mPackageClassMap);
  }
  
  /**
   * Finds all occurrences of the class in the class path.
   *
   * @param fullyQualifiedClassName the fully qualified class name to search
   * for in this classpath.
   * @return the list of ClassPathElements that the specified class is contained 
   * within.
   */
  public List findClass(String fullyQualifiedClassName) 
  {
    List classPathElementsThatContainClass = new ArrayList();
    for (Iterator i = mElements.iterator(); i.hasNext();) {
      ClassPathElement classPathElement = (ClassPathElement) i.next();
      if (classPathElement.containsClass(fullyQualifiedClassName)) {
        classPathElementsThatContainClass.add(classPathElement);
      }
    }
    return classPathElementsThatContainClass;
  }
  
  /**
   * Finds the fully qualified class names of the specified class name (unqualified).
   * For example, if rt.jar was in the classpath, a search for "List" would
   * return two items:
   *
   * 1) java.awt.List
   * 2) java.util.List
   *
   * @param unqualifiedClassName the classname to search for
   * @return a list of fully qualified classnames if any are found, otherwise
   * an empty list is returned.
   */
  public Set findUnqualifiedClass(String unqualifiedClassName)
  {
    Set packagesClassFoundIn = (Set) mClassPackageMap.get(unqualifiedClassName);
    Set fullyQualifiedClassNames = new HashSet();
    
    // iterate through package statements and add class name to each one
    if (packagesClassFoundIn != null) {
      for (Iterator i = packagesClassFoundIn.iterator(); i.hasNext();) {
        String currentPackage = (String) i.next();
        fullyQualifiedClassNames.add(currentPackage + "." + unqualifiedClassName);  
      }
    }
    
    return Collections.unmodifiableSet(fullyQualifiedClassNames);
  }
  
  /**
   * Finds any duplicate classes in this class path
   * 
   * @return a map containing the duplicate classes (if any).  The format of 
   * the map is as follows:
   *
   * KEY (Classname)  VALUE (Set of ClassPathElements where the class was found)
   * oracle.jbo.Row  [D:\source, D:\ctb\lib\crap.jar, etc...]
   * oracle.Shite    [...]
   * etc..
   */
  public Map findDuplicateClasses()
  {
    Map duplicateClassesMap = new HashMap();
    
    // for each element in the class path, find the intersection with every
    // other element in the class path to see if there is any overlap
    for (Iterator topIterator = mElements.iterator(); topIterator.hasNext();) {
      
      // get current top level element
      ClassPathElement topLevelElement = (ClassPathElement) topIterator.next();
      Set topLevelElementClasses = topLevelElement.getClasses();
      
      // compare this with every other element in the list
      List otherElements = new ArrayList(mElements);
      otherElements.remove(topLevelElement);
      for (Iterator j = otherElements.iterator(); j.hasNext();) {
        ClassPathElement otherElement = (ClassPathElement) j.next();
        
        // now get the intersection of the two
        Set intersection = new TreeSet(topLevelElementClasses);
        intersection.retainAll(otherElement.getClasses());

        // if we found some overlap add to map
        if (intersection.size() > 0) {
          for (Iterator k = intersection.iterator(); k.hasNext();) {
            String intersectingClass = (String) k.next();
            
            // see if a set has already been created, if not, then create it
            Set duplicateClassSet = (Set) duplicateClassesMap.get(intersectingClass);
            if (duplicateClassSet == null) {
              duplicateClassSet = new TreeSet();
              duplicateClassesMap.put(intersectingClass, duplicateClassSet);
            }
            
            // add BOTH the top level class path element and the other element
            // to the set
            duplicateClassSet.add(topLevelElement);
            duplicateClassSet.add(otherElement);
          }
        }
      }
      
    }
    
    return duplicateClassesMap;
  }

  /**
   * Gets all of the elements of this class path.
   * 
   * @return the elements of this class path
   */
  public List getElements()
  {
    if (mElements == null) {
      mElements = new ArrayList();
    }
    
    return Collections.unmodifiableList(mElements);
  }
  
  //
  // Private helper methods
  //
    
  /**
   * Update the package class map (package -> list of classes), by extracting 
   * the packages and associated classes from the specified class path element.
   *
   * @param element the ClassPathElement to extract the packages and classes
   * from.
   */
  private void updatePackageClassMap(ClassPathElement element)
  {
    // iterate through the packages adn their containing classes contained in
    // the specified ClassPathElement and add them the outer package class map.
    Map currentPackageClassMap = element.getPackageClassMap();
    for (Iterator i = currentPackageClassMap.keySet().iterator(); i.hasNext();) {
      String currentPackage = (String) i.next();
      
      // create list for this package if it doesn't already exist
      List currentPackageClassList = (List) mPackageClassMap.get(currentPackage);
      if (currentPackageClassList == null) {
        currentPackageClassList = new ArrayList();
        mPackageClassMap.put(currentPackage, currentPackageClassList);
      }
      
      currentPackageClassList.addAll((List)currentPackageClassMap.get(currentPackage));
    }
  }
  
  /**
   * Updates the class package map (class -> list of packages) with the 
   * specified ClassPathElement.
   * 
   * @param element the ClassPathElement to extract the packages and classes
   * from.
   */
  private void updateClassPackageMap(ClassPathElement element)
  {
    Map currentClassPackageMap = element.getClassPackageMap();
    
    for (Iterator i = currentClassPackageMap.keySet().iterator(); i.hasNext();) {
      String currentClass = (String) i.next();
      
      // create list for this class if it doesn't already exist
      Set packageSet = (Set) mClassPackageMap.get(currentClass);
      if (packageSet == null) {
        packageSet = new HashSet();
        mClassPackageMap.put(currentClass, packageSet);
      }
      
      packageSet.addAll((List)currentClassPackageMap.get(currentClass));
    }
   
  }
}
