/*
 * LazyImporter jEdit Plugin (ClassPathElement.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.etheridge.lazyimporter.classpath;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Represents an element in a classpath.  This will either be a 
 * jar file element (ie. D:\libraries\ant.jar) or a directory element
 * (ie. D:\classes).
 */
public abstract class ClassPathElement
{
  // the name of this class path element
  private String mName;
  
  // the element's position in the class path
  private int mPositionInClassPath;
  
  // a map that contains the classes inside this class path element
  // This map maps between a package name and a list of class names.
  // eg. package org.etheridge.lazyimporter.classpath maps to
  //       {ClassPath, ClassPathElement, DirElement, JarFileElement).
  protected Map mPackageClassMap;
  
  // mapping between classes and packages (inverse of above mapping)
  protected Map mClassPackageMap;
  
  // constant to signify that there is no package for a particular class.
  protected final static String NO_PACKAGE_MARKER = "NO_PACKAGE";
  
  /**
   * Constructs a ClassPathElement.
   *
   * @param name the name of this class path element.
   * @param position the position of this class path element in the
   * classpath.
   */
  public ClassPathElement(String name, int position)
  {
    mName = name;
    mPositionInClassPath = position;
  }
  
  /**
   * Abstract method to load the classes in this class path element.
   */
  protected abstract void loadClasses();
    
  /**
   * @return the name of this class path element. (eg. D:\temp\library.jar)
   */
  public String getName()
  {
    return mName;
  }
  
  /**
   * @return the mapping between the packages in the class path element and 
   * the associated classes within that package.
   */ 
  public Map getPackageClassMap()
  {
    // load the classes lazily
    if (mPackageClassMap == null) {
      loadClasses();
    }
    
    return Collections.unmodifiableMap(mPackageClassMap);
  }
  
  /**
   * @return the mapping between the classes in the class path element, and
   * the packages that each class is found within.
   */
  public Map getClassPackageMap()
  {
    // load the classes lazily
    if (mClassPackageMap == null) {
      loadClasses();
    }
    
    return Collections.unmodifiableMap(mClassPackageMap); 
  }
  
  /**
   * @return a set of strings representing the classes (well the classnames)
   * of the classes in this particular classpath element.
   */
  public Set getClasses()
  {
    // load the classes lazily
    if (mPackageClassMap == null) {
      loadClasses();
    }
    
    // go through the map and get the list of classes for each package
    Set fullyQualifiedClassSet = new HashSet();
    for (Iterator i = mPackageClassMap.keySet().iterator(); i.hasNext();) {
      String packageName = (String) i.next();
      List thisPackagesClasses = (List) mPackageClassMap.get(packageName);
      for (Iterator j = thisPackagesClasses.iterator(); j.hasNext();) {
        if (NO_PACKAGE_MARKER.equals(packageName)) {
          fullyQualifiedClassSet.add(j.next()); 
        } else {
          fullyQualifiedClassSet.add(packageName + "." + j.next());
        }
      }
    }
    
    return fullyQualifiedClassSet;
  }
  
  /**
   * Determines whether or not the named class is contained within this
   * class path element.
   * 
   * @param className the FULLY QUALIFIED name of the class to find
   * @return whether or not this class path element contains the specified
   * class.
   */
  public boolean containsClass(String fullyQualifiedClassName)
  {
    // separate the package and string
    int lastIndexOfDot = fullyQualifiedClassName.lastIndexOf(".");
    String packageName = NO_PACKAGE_MARKER;
    String className = null;
    if (lastIndexOfDot > 0) {
      packageName = fullyQualifiedClassName.substring(0, lastIndexOfDot);
      className = fullyQualifiedClassName.substring(lastIndexOfDot + 1);
    } else {
      className = fullyQualifiedClassName; 
    }
    
    // now attempt to find it the package in the map
    // lazily load package-class map
    if (mPackageClassMap == null) {
      loadClasses(); 
    }
    
    if (mPackageClassMap.containsKey(packageName)) {
      return ((List)mPackageClassMap.get(packageName)).contains(className); 
    } 
    
    return false;
  }
   
  /**
   * Overridden toString() method for debugging
   */ 
  public String toString()
  {
    StringBuffer buffer = new StringBuffer();
    
    // append name
    buffer.append("NAME: " + mName + "\n");
    
    // append classes in this jar file
    buffer.append("CLASSES: \n");
    
    // load classes if they haven't been loaded yet
    if (mPackageClassMap == null) {
      loadClasses();
    } 
    
    // print out classes
    for (Iterator i = getClasses().iterator(); i.hasNext();) {
      buffer.append(i.next() + "\n");
    }
       
    return buffer.toString();
  }
  
  public boolean equals(Object obj)
  {
    if (obj == null) {
      return false;
    }
    
    if (!(obj instanceof ClassPathElement)) {
      return false;
    }
    
    return getName().equalsIgnoreCase(((ClassPathElement)obj).getName());
  }
}
