/*
 * LazyImporter jEdit Plugin (ImportListWindow.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.etheridge.lazyimporter.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.KeyboardFocusManager;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;

import org.etheridge.lazyimporter.importer.ImportStatement;

/**
 * A window that contains a list of import statements, that can be chosen by
 * a user.  When the user chooses an import statement the window will close.
 * The window can also be closed by pressing the escape key.
 */
public class ImportListWindow extends JFrame
{
  private JList mImportList;
  private ImportListModel mImportListModel;
  private List mImportSelectionListeners;
   
  public ImportListWindow()
  {
    init();
  }
  
  /**
   * Sets the import statemetns to show in the window.
   *
   * @param importStatements an array of ImportStatements to show in the
   * window.
   */
  public void setImportStatements(ImportStatement[] importStatements)
  {
    // set the visible row count on the import list
    mImportList.setVisibleRowCount(importStatements.length);
    
    // clear the current model
    mImportListModel.clear();
    
    // add the new import statements to the model
    for (int i = 0; i < importStatements.length; i++) {
      mImportListModel.addElement(importStatements[i]);
    }
    
    mImportList.revalidate();
    mImportList.repaint();
  }
  
  //
  // Import Selection Listener Methods
  //
  
  public void addImportSelectionListener(ImportSelectionListener listener)
  {
    if (mImportSelectionListeners == null) {
      mImportSelectionListeners = new ArrayList();
    }
    mImportSelectionListeners.add(listener);
  }
  
  public void removeImportSelectionListener(ImportSelectionListener listener)
  {
    if (mImportSelectionListeners != null) {
      mImportSelectionListeners.remove(listener);
    }
  }
  
  /**
   * Overridden to get around focusing problem!
   * NOTE: if anyone knows how to get around this bloody problem, let me know! ;)
   */
  public void setVisible(boolean visible)
  {
    super.setVisible(visible);
    
    // select first item in list
    mImportList.setSelectedIndex(0);
    
    // move the window to the front
    toFront();
    
    // let the import list have focus
    mImportList.requestFocus();
  }
  
  //
  // Private Helper Methods
  //
  
  private void init()
  {
    getContentPane().setLayout(new BorderLayout());

    // make this JFrame undecorated (ie. remove funky top-right stuff)
    setUndecorated(true);
    
    // create import list
    mImportListModel = new ImportListModel();
    mImportList = new JList(mImportListModel);
    mImportList.setBorder(BorderFactory.createEtchedBorder());
    
    // escape listener (to close the window when escape is pressed).
    KeyStroke ks = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0, false);
    mImportList.getInputMap().put(ks, ks);
    mImportList.getActionMap().put(ks, new AbstractAction() 
    {
      public void actionPerformed(ActionEvent ae) 
      {
        closeWindow();
      }
    });
    
    // enter listener (to signify a user pressed enter to select an import statement)
    KeyStroke enterKeyStroke = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false);
    mImportList.getInputMap().put(enterKeyStroke, enterKeyStroke);
    mImportList.getActionMap().put(enterKeyStroke, new AbstractAction() 
    {
      public void actionPerformed(ActionEvent ae) 
      {
        selectionMade();
      }
    });

    // when the import list loses focus, remove the containing import list
    // window.
    mImportList.addFocusListener(new FocusAdapter()
    {
      public void focusLost(FocusEvent e)
      {
        // when the import list loses focus, we want to make sure the window
        // is closed.
        if (KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusedWindow() != ImportListWindow.this) {
          closeWindow();
        }
      }
    });

    // add a mouse listener for user selection
    mImportList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    mImportList.addMouseListener(new MouseAdapter()
    {
      public void mouseClicked(MouseEvent e)
      {
        // double-click signifies a user selection
        if (e.getClickCount() >= 2) {
          selectionMade();
        }
      }
    });
        
    getContentPane().add(new JScrollPane(mImportList));
    pack();
  }

  /**
   * Notify any import selection listeners that an import statement was 
   * selected.
   */
  private void notifyImportSelectionListeners(ImportStatement statement)
  {
    if (mImportSelectionListeners != null) {
      for (Iterator i = mImportSelectionListeners.iterator(); i.hasNext();) {
        ImportSelectionListener listener = (ImportSelectionListener) i.next();
        listener.importSelected(statement);
      }
    }
  }
  
  private void selectionMade()
  {
    notifyImportSelectionListeners((ImportStatement)mImportList.getSelectedValue());
    closeWindow();
  }
 
  private void closeWindow()
  {
    hide();
  }
  
}
