/*
 * LazyImporter jEdit Plugin (JavaClass.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.etheridge.lazyimporter.importer;

import gnu.regexp.RE;
import gnu.regexp.REException;
import gnu.regexp.REMatch;
import gnu.regexp.RESyntax;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import org.etheridge.lazyimporter.classpath.ClassPath;
import org.etheridge.lazyimporter.importer.ClassImportStatement;
import org.etheridge.lazyimporter.importer.drawing.DefaultDrawingStrategy;
import org.etheridge.lazyimporter.importer.drawing.DrawingStrategy;
import org.etheridge.lazyimporter.importer.ImportStatement;
import org.etheridge.lazyimporter.importer.ordering.DefaultOrderingStrategy;
import org.etheridge.lazyimporter.importer.ordering.OrderingStrategy;
import org.etheridge.lazyimporter.importer.PackageImportStatement;
import org.etheridge.lazyimporter.utility.JavaUtilities;

/**
 * This class represents a coded version of a Java Class (ie. the text that
 * makes up a java class). 
 */
public class JavaClass
{
  // the text of the class
  private String mClassText;
  
  // the text of the class without the comments
  private String mClassTextMinusComments;
 
  // this class's name
  private String mClassName;
 
  // list of import statements (ImportStatement objects)
  private List mImportStatements;
  
  // the set of classes explicity imported
  private Set mImportedClasses;
  
  // implicit import statement for this class, determined by the package
  // this class is in (ie. from the package statement).
  private PackageImportStatement mThisPackageImportStatement;
  
  // the ordering strategies - order of strategy is important! ;)
  private List mOrderingStrategies;
  
  // the drawing strategy
  private DrawingStrategy mDrawingStrategy;
  
  // regular expression strings
  private static final String msAllImportsRegularExpression 
    = "import[ ]+[[:alpha:]][[:alnum:]_]*([\\.][[:alpha:]][[:alnum:]_]*)*(\\.\\*)?[ ]*;";
  private static final String msPackageStatementRegularExpression  
    = "[^][[:space:]]*package[[:space:]]+[[:alnum:].$_*]*;";
  private static final String msClassNameRegularExpression
    = "[ ](class|interface)[ ]+[[:alpha:]][[:alnum:]_]*";

  // static strings
  private static String msValidPreClassCharacters = "\t() \n,!";
  private static String msValidPostClassCharacters = "\t)( .\n[,;";
  
  public JavaClass(String classText)
  {
    mClassText = classText;
    mClassTextMinusComments = JavaUtilities.removeStringLiterals(JavaUtilities.stripJavaComments(classText));
    
    // set the ordering strategy to be the default one
    mOrderingStrategies = new ArrayList();
    mOrderingStrategies.add(new DefaultOrderingStrategy());
    
    // set the drawing strategy to be the default one
    mDrawingStrategy = new DefaultDrawingStrategy();
    
    // find and store this java classe's name
    findClassName();
  }
	 
  /**
   * @return the text of this class.
   */ 
  public String getClassText()
  {
    return mClassText;
  }
  
  /** 
   * @return this classes implicit package import statement.  Returns null if
   * this class has no package statement.
   */
  public ImportStatement getImplicitImportStatement()
  {
    return mThisPackageImportStatement;
  }
  
  /**
   * Add the specified import to this java class.  This will be added after
   * all other imports.
   * 
   * @param importStatement the import statement to add
   */
  public void addImportStatement(ImportStatement importStatement)
  {
    if (mImportStatements == null) {
      findImports();
    }
    
    // add it to the end
    addImportStatement(importStatement, -1);
  }
  
  /**
   * Adds the specified import statement, overwriting any existing import
   * statements that have the same class name.
   *
   * @param importStatemetn the import statement to add.
   */
  public void overwriteImportStatement(ImportStatement importStatement)
  {
    if (mImportStatements == null) {
      findImports();
    }
    
    // see if there are any existing import statements with the same class
    // name and remove it.
    if (importStatement instanceof ClassImportStatement) {
      ClassImportStatement classImportStatement = (ClassImportStatement) importStatement;
      if (mImportedClasses.contains(classImportStatement.getClassName())) {
        // find import and remove it
        for (Iterator i = mImportStatements.iterator(); i.hasNext();) {
          ImportStatement currentImportStatement = (ImportStatement) i.next();
          if (currentImportStatement instanceof ClassImportStatement) {
            if (((ClassImportStatement)currentImportStatement).getClassName().equals(classImportStatement.getClassName())) {
              // remove from import statement list
              i.remove();
              
              // remove from classes set
              mImportedClasses.remove(((ClassImportStatement)currentImportStatement).getClassName());
              
            }
          }
        }
      }
    }
    
    // add it to the end
    addImportStatement(importStatement, -1);
  }
  
  
  /** 
   * Sets the ordering strategies to be used to display the import statements.
   * 
   * @param orderingStrategies an array of ordering strategies to be used. The
   * order of the strategies in the array defines the order in which the 
   * ordering strategies are applied. (if that makes sense!) ;)
   */
  public void setOrderingStrategies(OrderingStrategy[] orderingStrategies)
  {
    mOrderingStrategies.clear();
    mOrderingStrategies.addAll(Arrays.asList(orderingStrategies));
  }
  
  public List getOrderingStrategies()
  {
    return Collections.unmodifiableList(mOrderingStrategies);
  }
  
  public void setDrawingStrategy(DrawingStrategy drawingStrategy)
  {
    mDrawingStrategy = drawingStrategy;
  }
  
  /**
   * @return a list of all import statements of this class as they currently
   * stand.
   */
  public List getImportStatements()
  {
    // lazily find imports
    if (mImportStatements == null) {
      findImports();
    }
    return Collections.unmodifiableList(mImportStatements);
  }
  
  /**
   * Removes any unused imports from this java class.  Subsequent calls to 
   * getImportStatements() will result in a list of ImportStatements that does
   * *NOT* contain any unused imports.
   */
  public void removeUnusedImports()
  {
    // lazily find imports
    if (mImportStatements == null) {
      findImports();
    }
    
    // iterate through import statements, and see if the class is actually used
    // in this java class code
    for (Iterator i = mImportStatements.iterator(); i.hasNext();) {
      ImportStatement importStatement = (ImportStatement) i.next();
      if (importStatement instanceof ClassImportStatement) {
        String className = ((ClassImportStatement)importStatement).getClassName();
        boolean foundClass = findClassUsage(className);
        if (!findClassUsage(className)) {
          i.remove();
        }
      }
    }
  }
  
  /** 
   * Gets the import statement at the specified position in this java class.
   * @return the import statement at the specified position in this java
   * class.
   */
  public ImportStatement getImportStatementAtPosition(int position)
  {
    try {
      RE allImportsRegularExpression = new RE(msAllImportsRegularExpression, RE.REG_MULTILINE, RESyntax.RE_SYNTAX_POSIX_EXTENDED);
      REMatch[] importMatches = allImportsRegularExpression.getAllMatches(mClassTextMinusComments);
      for (int i = 0; i < importMatches.length; i++) {
        int startIndex = importMatches[i].getStartIndex();
        int endIndex = importMatches[i].getEndIndex();
        if (position >= startIndex && position <= endIndex) {
          String classImportStatement = mClassTextMinusComments.substring(startIndex, endIndex);
          return ImportStatement.create(classImportStatement);
        }
      }
    } catch (REException reException) {
      reException.printStackTrace();
    }
    return null;
  }
  
  /**
   * Converts all explicit imports to wildcard (package) imports.
   *
   * eg. java.util.Date -> java.util.*;
   */
  public void convertExplicitImportsToPackageImports()
  {
    // lazily find imports
    if (mImportStatements == null) {
      findImports();
    }
    
    // iterate through all of the import statements, and convert any explicit
    // imports into wildcard imports.
    Set packagesImported = new HashSet();
    List wildcardImportList = new ArrayList();
    for (Iterator i = mImportStatements.iterator(); i.hasNext();) {
      ImportStatement currentImportStatement = (ImportStatement) i.next();
      String currentPackageName = currentImportStatement.getPackageName();
      
      // if this package hasn't been imported yet, add it if it's a wildcard
      // and change to a wildcard if it's explicit.
      if (!packagesImported.contains(currentPackageName)) {
        ImportStatement importStatementToAdd = currentImportStatement;
        if (currentImportStatement instanceof ClassImportStatement) {
          importStatementToAdd = ImportStatement.create(currentPackageName + ".*");
        } 
        wildcardImportList.add(importStatementToAdd);
        packagesImported.add(currentPackageName);
      }
    }
    
    mImportStatements.clear();
    mImportStatements.addAll(wildcardImportList);
  }
  
  /**
   * Resolves any package (wildcard) import statements using the specified 
   * class path to search for classes.
   *
   * @param classPath this classpath is used to search for all classes within
   * each wildcard package so class usage can be resolved.
   */
  public void resolvePackageImports(ClassPath classPath, boolean resolveImportsInThisPackage)
  {
    // lazily find imports
    if (mImportStatements == null) {
      findImports();
    }
        
    // find all the fully qualified imports and add to list
    List fullyQualifiedImports = new ArrayList();
    for (Iterator i = mImportStatements.iterator(); i.hasNext();) {
      ImportStatement importStatement = (ImportStatement) i.next();
      if (importStatement instanceof ClassImportStatement) {
        fullyQualifiedImports.add(importStatement);
      }
    }
    
    // for each of the package import statements attempt to resolve the classes
    List packageImportStatements = new ArrayList(mImportStatements);
    packageImportStatements.removeAll(fullyQualifiedImports);
    
    // if we want to resolve this packages imports explicitly then add them to
    // the list to the END of this list to resolve.
    if (resolveImportsInThisPackage && mThisPackageImportStatement != null) {
      packageImportStatements.add(mThisPackageImportStatement);
    }
    
    for (Iterator i = packageImportStatements.iterator(); i.hasNext();) {
      PackageImportStatement packageImportStatement = (PackageImportStatement) i.next();
      String currentPackageString = packageImportStatement.getPackageName();
      if (classPath.getPackageClassMap().containsKey(currentPackageString)) {
        // For every class in this package, see if the class is used in this
        // class.  If it is, then create a new import statement for it.
        List classesInPackage = (List) classPath.getPackageClassMap().get(currentPackageString);
        for (Iterator classIterator = classesInPackage.iterator(); classIterator.hasNext();) {
          String currentClass = (String) classIterator.next();
          // if a class usage is found, add a new CLASS import statement if one
          // does not already exist for it.
          if (findClassUsage(currentClass)) {
            ClassImportStatement newImportStatement = new ClassImportStatement(currentPackageString, currentClass);
            if (!mImportStatements.contains(newImportStatement)) {
              addImportStatement(newImportStatement, mImportStatements.indexOf(packageImportStatement));
            }
          }
        }
        
        // remove the package import statement, as it is no longer required.
        mImportStatements.remove(packageImportStatement);
      }
    }
  }
  
  public void redrawImportStatements()
  {
    // lazily find imports
    if (mImportStatements == null) {
      findImports();
    }
    
    // if there are no import statements to print, then just return
    if (mImportStatements == null) {
      return;
    }
    
    int firstImportStart = removeCurrentImports();

    // reorder the list of according to the ordering strategies defined
    List reorderedImportStatements = new ArrayList(mImportStatements);
    for (Iterator i = mOrderingStrategies.iterator(); i.hasNext();) {
      reorderedImportStatements = 
        ((OrderingStrategy)i.next()).orderImports(this, reorderedImportStatements);
    }
 
    // use drawing strategy to get visual representation of imports
    String drawnImports = mDrawingStrategy.drawImports(this, reorderedImportStatements);
    
    // redraw list of import statements
    StringBuffer buffer = new StringBuffer(mClassText);
    buffer.insert(firstImportStart, drawnImports);
    mClassText = buffer.toString();
    
    StringBuffer minusCommentsBuffer = new StringBuffer(mClassTextMinusComments);
    minusCommentsBuffer.insert(firstImportStart, drawnImports);
    mClassTextMinusComments = minusCommentsBuffer.toString();
  }
     
  //
  // private helper methods
  //

  private int removeCurrentImports()
  {
    // if there is a current import statemetn, then this is where we will 
    // insert the new import statements.
    try {
      RE allImportsRegularExpression = new RE(msAllImportsRegularExpression, RE.REG_MULTILINE, RESyntax.RE_SYNTAX_POSIX_EXTENDED);
      REMatch[] importMatches = allImportsRegularExpression.getAllMatches(mClassTextMinusComments);
      if (importMatches.length > 0) {
        int firstImportStart = importMatches[0].getStartIndex();
        int lastImportEnd = importMatches[importMatches.length - 1].getEndIndex();
        StringBuffer buffer = new StringBuffer(mClassText);
        StringBuffer bufferMinusComments = new StringBuffer(mClassTextMinusComments);
        mClassText = buffer.delete(firstImportStart, lastImportEnd).toString();
        mClassTextMinusComments = bufferMinusComments.delete(firstImportStart, lastImportEnd).toString();
        return firstImportStart;  
      }
  
      // otherwise look for package statement
      RE packageStatementRegularExpression = new RE(msPackageStatementRegularExpression, RE.REG_MULTILINE, RESyntax.RE_SYNTAX_POSIX_EXTENDED);
      REMatch packageMatch = packageStatementRegularExpression.getMatch(mClassTextMinusComments);
      if (packageMatch != null) {
        return packageMatch.getEndIndex() + 1;
      }
    } catch (REException reException) {
      // should never get here
      reException.printStackTrace();
    }
    
    // if there is no imports or package statement, just return 0; (beginning)
    return 0;
  }
  
  private void findImports()
  {
    mImportStatements = new ArrayList();
    mImportedClasses = new HashSet();
        
    try {
      // find all import statements
      RE allImportsRegularExpression = new RE(msAllImportsRegularExpression, RE.REG_MULTILINE, RESyntax.RE_SYNTAX_POSIX_EXTENDED);
      REMatch[] importMatches = allImportsRegularExpression.getAllMatches(mClassTextMinusComments);
      for (int i = 0; i < importMatches.length; i++) {
        int startIndex = importMatches[i].getStartIndex();
        int endIndex = importMatches[i].getEndIndex();
        String classImportStatement = mClassTextMinusComments.substring(startIndex, endIndex);
        addImportStatement(ImportStatement.create(classImportStatement), -1);
      }
  
      // find package statement (if it exists)
      RE packageStatementRegularExpression = new RE(msPackageStatementRegularExpression, RE.REG_MULTILINE, RESyntax.RE_SYNTAX_POSIX_EXTENDED);
      REMatch packageMatch = packageStatementRegularExpression.getMatch(mClassTextMinusComments);
      if (packageMatch != null) {
        int startIndex = packageMatch.getStartIndex();
        int endIndex = packageMatch.getEndIndex();
        String packageStatement = mClassTextMinusComments.substring(startIndex, endIndex);
        mThisPackageImportStatement = 
          (PackageImportStatement) ImportStatement.create(packageStatement); 
      }
    } catch (REException reException) {
      // this should NEVER happen!
      reException.printStackTrace();
    }
  }
    
  /**
   * @return whether or not the named class is found in this Java class code.
   * 
   * NOTE: this was implemented using a dynamic regular expression, however it
   * was WAY too slow!  This may be uglier, but it is much quicker.
   */
  public boolean findClassUsage(String className)
  {
    // if the class name we are looking for is the same as the current class,
    // then just return false as we do *NOT* want to include an import 
    // statement for the current class.
    if (className.equals(mClassName)) {
      return false;
    }
    
    // tokenize on new lines
    StringTokenizer tokenizer = new StringTokenizer(mClassTextMinusComments, "\n");
    while (tokenizer.hasMoreTokens()) {
      String currentLine = tokenizer.nextToken();
      // find any occurrences of this class name in this line, making sure that
      // the class name has valid characters surrounding it.
      int index = -1;
      while ((index = currentLine.indexOf(className)) >= 0) {
        boolean validPreClassCharacter = true;
        boolean validPostClassCharacter = true;
        
        // check pre classname character if not at the beginning of the line
        if (index > 0) { 
          validPreClassCharacter = isValidPreClassCharacter(currentLine.charAt(index - 1));
        }  
        
        // check post classname character if not at end of line
        if ((className.length() + index) < currentLine.length()) {
          validPostClassCharacter = 
            isValidPostClassCharacter(currentLine.charAt(className.length() + index));
        }
        
        if (validPreClassCharacter && validPostClassCharacter) {
          return true;
        }
        
        currentLine = currentLine.substring(index + 1);
      }
    }
     
    return false;
  }

  /**
   * @return whether the specified character can be before a valid class usage.
   */ 
  private boolean isValidPreClassCharacter(char ch)
  {
    return msValidPreClassCharacters.indexOf(ch) >= 0;
  }
  
  /**
   * @return whether the specified character can be after a valid class usage.
   */ 
  private boolean isValidPostClassCharacter(char ch)
  {
    return msValidPostClassCharacters.indexOf(ch) >= 0;
  }
  
  /**
   * Adds an import statement to the list of import statements, making sure 
   * any explicit class imports are not already imported.
   * 
   * @param importStatement the import statement to add.
   * @param index the index to add the import statement in.  If this is -1,
   * the statement is added to the end of the import statement list.
   */ 
  private void addImportStatement(ImportStatement importStatement, int index)
  {
    // if this is an implicit class import statement, only add it if the class
    // has not already been imported.
    if (importStatement instanceof ClassImportStatement) {
      String className = ((ClassImportStatement)importStatement).getClassName();
      if (mImportedClasses.contains(className)) {
        return;
      }
      
      // if the class being added is actually an import statemetn for THIS class
      // then do not add it and return (as this causes compile errors in 1.4 onwards)
      if (isImportStatementForThisClass((ClassImportStatement)importStatement)) {
        return;
      }
            
      // add to set of implicity loaded classes
      mImportedClasses.add(className);
    }
        
    if (index != -1) {
      mImportStatements.add(index, importStatement);
    } else {
      mImportStatements.add(importStatement);
    }
  }

  /**
   * Find and store this class's outermost class/interface name (first one in file)
   */
  private void findClassName()
  {
    try {
      RE classRegularExpression = new RE(msClassNameRegularExpression, RE.REG_MULTILINE, RESyntax.RE_SYNTAX_POSIX_EXTENDED);
      REMatch classMatch = classRegularExpression.getMatch(mClassTextMinusComments);
      if (classMatch != null) {
        int startIndex = classMatch.getStartIndex();
        int endIndex = classMatch.getEndIndex();
        String classStatement = mClassTextMinusComments.substring(startIndex, endIndex);
        
        // if this is a class then remove the class from teh string
        if (classStatement.indexOf("class") >= 0) {
          int classKeywordIndex = classStatement.indexOf("class");
          mClassName = classStatement.substring(classKeywordIndex + 5).trim();  
        } else { // must be an interface, so remove interface - hasty assumption?
          int classKeywordIndex = classStatement.indexOf("interface");
          mClassName = classStatement.substring(classKeywordIndex + 9).trim();
        }
      }
    } catch (REException reException) {
      // this shouldn't happen, so just print the stack trace
      reException.printStackTrace();
    }
  }
  
  /**
   * @return whether or not the specified import statement is an import
   * statement for this class!
   */
  private boolean isImportStatementForThisClass(ClassImportStatement classImportStatement)
  {
    if (mThisPackageImportStatement == null) {
      return false;
    }
    
    String thisClassesPackage = mThisPackageImportStatement.getPackageName();
    String comparePackage = classImportStatement.getPackageName();
    
    // if the packages are not the same, then there is no way this is a match
    if (!comparePackage.equals(thisClassesPackage)) {
      return false;
    }
    
    // packages must be the same, so see if the classes are the same
    return (mClassName.equals(classImportStatement.getClassName()));
  }
}
