/*
 * LazyImporter jEdit Plugin (GroupingDrawingStrategy.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.etheridge.lazyimporter.importer.drawing;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.etheridge.lazyimporter.importer.ImportStatement;
import org.etheridge.lazyimporter.importer.JavaClass;
import org.etheridge.lazyimporter.importer.ordering.AlphabeticalOrderingStrategy;
import org.etheridge.lazyimporter.importer.ordering.OrderingStrategy;

/**
 * Draws the ImportStatements in "groups" - these groups of import statements
 * are broken up according to the "level" of the import statement.  In other
 * words, how many "dots" of the import statement you want to group together.
 * (starting from the left-hand side).
 */
public class GroupingDrawingStrategy implements DrawingStrategy
{
  public static final int DEFAULT_GROUP_LEVEL = 1;
  public static final int DEFAULT_GROUP_SPACING = 1;
  
  // the level of grouping
  private int mGroupingLevel;
  
  // the number of newlines between groups
  private int mGroupingSpace;
  
  /**
   * Constructor.
   *
   * @param groupingLevel the "level" to group the imports at.
   * @param groupingSpace the number of lines between groups.
   */
  public GroupingDrawingStrategy(int groupingLevel, int groupingSpace)
  {
    mGroupingLevel = groupingLevel;
    mGroupingSpace = groupingSpace;
  }
  
  public String drawImports(JavaClass javaClass, List importStatements)
  {
    StringBuffer buffer = new StringBuffer();
    
    // get the current ordering strategies from the JavaClass and make sure the
    // alphabetical ordering strategy is at the beginning.
    List orderingStrategies = new ArrayList(javaClass.getOrderingStrategies());
    for (Iterator i = orderingStrategies.iterator(); i.hasNext();) {
      OrderingStrategy currentStrategy = (OrderingStrategy) i.next();
      if (currentStrategy instanceof AlphabeticalOrderingStrategy) {
        i.remove();
      }
    }
    orderingStrategies.add(0, new AlphabeticalOrderingStrategy());
    
    // reapply the ordering strategies to the list of imports
    // (yes, I know this sucks, as we are doing it again, but it'll do for now).
    List reorderedImportStatements = new ArrayList(importStatements);
    for (Iterator i = orderingStrategies.iterator(); i.hasNext();) {
      OrderingStrategy orderingStrategy = (OrderingStrategy) i.next();
      reorderedImportStatements = 
        orderingStrategy.orderImports(javaClass, reorderedImportStatements);
    }
    
    // for each import statement, get the grouping part of it (according to grouping
    // level) and add to map for grouping.
    Map groupingMap = new HashMap(); // groups "grouping parts" to a List of ImportStatements
    List orderedPartList = new ArrayList();
    for (Iterator i = reorderedImportStatements.iterator(); i.hasNext();) {
      ImportStatement currentStatement = (ImportStatement) i.next();
      String groupingPart = getGroupingImportPart(currentStatement);
      
      // add to hashmap, keyed on the grouping part
      List currentGroupingList = (List) groupingMap.get(groupingPart);
      if (currentGroupingList == null) {
        currentGroupingList = new ArrayList();
        groupingMap.put(groupingPart, currentGroupingList);
        orderedPartList.add(groupingPart); // add to this list to keep key ordering
      }
      currentGroupingList.add(currentStatement);
    }
    
    // grab stuff out of the map and add to buffer
    for (Iterator i = orderedPartList.iterator(); i.hasNext();) {
      String currentKey = (String) i.next();
      List groupedImportStatements = (List) groupingMap.get(currentKey);
      for (Iterator j = groupedImportStatements.iterator(); j.hasNext();) {
        ImportStatement currentImportStatement = (ImportStatement) j.next();
        buffer.append(currentImportStatement.getActualImportStatement() + ";");
        if (j.hasNext()) {
          buffer.append("\n");
        }
      }
      
      // add spacing string if there are more groups to print
      if (i.hasNext()) {
        for (int counter = 0; counter <= mGroupingSpace; counter++) {
          buffer.append("\n");
        }
      }
      
    }
    return buffer.toString();
  }
  
  private String getGroupingImportPart(ImportStatement importStatement)
  {
    String packageName = importStatement.getPackageName();
    StringTokenizer tokenizer = new StringTokenizer(packageName, ".");
    StringBuffer groupingPartBuffer = new StringBuffer();
    
    int levelCounter = 0;
    while (tokenizer.hasMoreTokens() && levelCounter < mGroupingLevel) {
      groupingPartBuffer.append(tokenizer.nextToken());
      levelCounter++;
      if (tokenizer.hasMoreTokens() && levelCounter < mGroupingLevel) {
        groupingPartBuffer.append(".");
      }
    }
    
    return groupingPartBuffer.toString();
  }
}
