/*
 * LazyImporter jEdit Plugin (LayoutOptionsPane.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.etheridge.lazyimporter.options;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;

import java.lang.Object;
import java.lang.String;

import java.util.Iterator;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.ListSelectionModel;

import org.etheridge.lazyimporter.importer.drawing.GroupingDrawingStrategy;
import org.etheridge.lazyimporter.importer.grouping.AllRemainingImportGroupElement;
import org.etheridge.lazyimporter.importer.grouping.BlankLineImportGroupElement;
import org.etheridge.lazyimporter.importer.grouping.*;
import org.etheridge.lazyimporter.importer.grouping.ImportGroupElementFactory;
import org.etheridge.lazyimporter.importer.grouping.PrefixImportGroupElement;
import org.etheridge.lazyimporter.importer.JavaClass;
import org.etheridge.lazyimporter.LazyImporterProperties;

import org.gjt.sp.jedit.AbstractOptionPane;
import org.gjt.sp.jedit.jEdit;


/**
 * LazyImporter layout options (ie. drawing options - how the improts will be
 * layed out/drawn back into the java class text).  
 *
 * This options pane contains a preview panel, which will react dynamically to 
 * changes in the grouping level and grouping spaces user settings.
 */
public class LayoutOptionsPane extends AbstractOptionPane implements ActionListener
{
  private JCheckBox mGroupImportsCheckBox;
  private JSlider mGroupLevelSlider;
  private JSlider mGroupSpacingSlider;
  private LayoutPreviewer mLayoutPreviewer;
  
  private JCheckBox mManualGroupingCheckBox;
  private ImportGroupList mImportGroupList;
  private JButton mAddButton;
  private JButton mRemoveButton;
  private JButton mUpButton;
  private JButton mDownButton;
  private JButton mAddSpacerButton;
  private JButton mAddCommentButton;
    
  private JTabbedPane mTabbedPane;
  
  private static final String msPreviewString = "package this.is.package.statement;\n\nimport java.util.List;\nimport java.util.ArrayList;\nimport javax.swing.JTable;\nimport javax.swing.JList;\nimport org.etheridge.lazyimporter.LazyImporterPlugin;\nimport javax.swing.event.ChangeEvent;\nimport javax.swing.event.ChangeListener;\nimport javax.naming.Context;\nimport javax.rmi.PortableRemoteObject;";
    
  public LayoutOptionsPane()
  {
    super("LazyImporter.LayoutOptionPane");
  }

	public void _init()
  {
    // top level border layout
    setLayout(new BorderLayout());
    
    // create tabbed pane
    mTabbedPane = new JTabbedPane();
    
    // create automatic grouping panel
    JPanel autoGroupingPanel = new JPanel(new BorderLayout(6,6));
    autoGroupingPanel.add(createGroupingPanel(), BorderLayout.NORTH);
    autoGroupingPanel.add(createPreviewPanel(), BorderLayout.CENTER);
    
    // create manual grouping panel
    JPanel manualGroupingPanel = new JPanel(new BorderLayout(6,6));
    manualGroupingPanel.add(createManualGroupingPanel(), BorderLayout.NORTH);
    
    // add to tabbed pane
    mTabbedPane.add("Automatic", autoGroupingPanel);
    mTabbedPane.add("Manual", manualGroupingPanel);
    add(mTabbedPane, BorderLayout.CENTER);
    
    // determine which tab to show
    mTabbedPane.setSelectedComponent(jEdit.getBooleanProperty
      (LazyImporterProperties.MANUAL_GROUP_IMPORTS, false) ? manualGroupingPanel : autoGroupingPanel);
  }
   
  public void _save()
  {
    setProperties();
  }
  
  //
  // GUI Initialization
  //
  
  private JPanel createGroupingPanel()
  {
    JPanel groupingPanel = new JPanel(new BorderLayout(6,6));
    groupingPanel.setBorder(BorderFactory.createCompoundBorder
      (BorderFactory.createTitledBorder(jEdit.getProperty("LazyImporter.LayoutOptionPane.Grouping.Title")), BorderFactory.createEmptyBorder(6,6,6,6)));

    mGroupImportsCheckBox = new JCheckBox(jEdit.getProperty("LazyImporter.LayoutOptionPane.GroupImports"));
    mGroupImportsCheckBox.setSelected(jEdit.getBooleanProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS, false));
    mGroupImportsCheckBox.addChangeListener(new ChangeListener()
    {
      public void stateChanged(ChangeEvent e)
      {
        mGroupLevelSlider.setEnabled(mGroupImportsCheckBox.isSelected());
        mGroupSpacingSlider.setEnabled(mGroupImportsCheckBox.isSelected());
        mLayoutPreviewer.setEnabled(mGroupImportsCheckBox.isSelected());
        
        // make sure the manual gropuing checkbox is not selected if this is
        // selected
        if (mGroupImportsCheckBox.isSelected()) {
          mManualGroupingCheckBox.setSelected(false);
        }
      }
    });

    // create sliders
    mGroupLevelSlider = new JSlider(JSlider.HORIZONTAL, 0, 10, 1);
    mGroupLevelSlider.setMajorTickSpacing(2);
    mGroupLevelSlider.setMinorTickSpacing(1);
    mGroupLevelSlider.setPaintTicks(true);
    mGroupLevelSlider.setPaintLabels(true);
    mGroupLevelSlider.setSnapToTicks(true);
    mGroupLevelSlider.setValue(jEdit.getIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_LEVEL, 1));
    mGroupLevelSlider.setEnabled(mGroupImportsCheckBox.isSelected());
    
    mGroupSpacingSlider = new JSlider(JSlider.HORIZONTAL, 0, 10, 1);
    mGroupSpacingSlider.setMajorTickSpacing(2);
    mGroupSpacingSlider.setMinorTickSpacing(1);
    mGroupSpacingSlider.setPaintTicks(true);
    mGroupSpacingSlider.setPaintLabels(true);
    mGroupSpacingSlider.setSnapToTicks(true);
    mGroupSpacingSlider.setValue(jEdit.getIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_SPACING, 1));
    mGroupSpacingSlider.setEnabled(mGroupImportsCheckBox.isSelected());

    // add a change listener for each slider
    ChangeListener changeListener = new GroupLayoutChangeListener();
    mGroupLevelSlider.addChangeListener(changeListener);
    mGroupSpacingSlider.addChangeListener(changeListener);
        
    GridBagLayout gridBagLayout = new GridBagLayout();
    JPanel sliderPanel = new JPanel(gridBagLayout);
    GridBagConstraints cons = new GridBagConstraints();
		cons.gridx = 0;
    cons.gridy = 0;
		cons.gridheight = 1;
		cons.gridwidth = 1;
		cons.weightx = 0.0f;
		cons.insets = new Insets(1,0,1,0);
		cons.fill = GridBagConstraints.WEST;
		cons.anchor = GridBagConstraints.EAST;
    JLabel groupLevelLabel = new JLabel(jEdit.getProperty("LazyImporter.LayoutOptionPane.GroupLevel"));
		sliderPanel.add(groupLevelLabel, cons);
    
		cons.gridx = 1;
		cons.weightx = 1.0f;
    cons.fill = GridBagConstraints.HORIZONTAL;
    cons.gridwidth = GridBagConstraints.REMAINDER;
		gridBagLayout.setConstraints(mGroupLevelSlider,cons);
		sliderPanel.add(mGroupLevelSlider);
           
		cons.gridx = 0;
    cons.gridy = 1;
		cons.gridheight = 1;
		cons.gridwidth = 1;
		cons.weightx = 0.0f;
		cons.insets = new Insets(1,0,1,0);
		JLabel groupSpacingLabel = new JLabel(jEdit.getProperty("LazyImporter.LayoutOptionPane.GroupSpacing"));
    gridBagLayout.setConstraints(groupSpacingLabel,cons);
		sliderPanel.add(groupSpacingLabel);
    
    cons.gridx = 1;
		cons.weightx = 1.0f;
		gridBagLayout.setConstraints(mGroupSpacingSlider,cons);
		sliderPanel.add(mGroupSpacingSlider);

    // add to panel
    groupingPanel.add(mGroupImportsCheckBox, BorderLayout.NORTH);
    groupingPanel.add(sliderPanel, BorderLayout.CENTER);
        
    return groupingPanel;
  }
  
  private JPanel createPreviewPanel()
  {
    JPanel previewPanel = new JPanel(new BorderLayout(6,6));
    previewPanel.setBorder(BorderFactory.createCompoundBorder
      (BorderFactory.createTitledBorder(jEdit.getProperty("LazyImporter.LayoutOptionPane.Preview.Title")), BorderFactory.createEmptyBorder(6,6,6,6)));
    
    mLayoutPreviewer = new LayoutPreviewer
      (jEdit.getIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_LEVEL, 1),
       jEdit.getIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_SPACING, 1));
    
    previewPanel.add(mLayoutPreviewer, BorderLayout.CENTER);
    return previewPanel;
  }
  
  // manual grouping panel
  
  private JPanel createManualGroupingPanel()
  {
    JPanel groupingPanel = new JPanel(new BorderLayout(6,6));
    groupingPanel.setBorder(BorderFactory.createCompoundBorder
      (BorderFactory.createTitledBorder("Manual Grouping"), BorderFactory.createEmptyBorder(6,6,6,6)));
    
    mManualGroupingCheckBox = new JCheckBox("Manually group imports?");
    mManualGroupingCheckBox.setSelected(jEdit.getBooleanProperty(LazyImporterProperties.MANUAL_GROUP_IMPORTS, false));
    mManualGroupingCheckBox.addChangeListener(new ChangeListener()
    {
      public void stateChanged(ChangeEvent e)
      {
        // make sure the automatic group imports checkbox is not selected if
        // this checkbox is selected
        if (mManualGroupingCheckBox.isSelected()) {
          mGroupImportsCheckBox.setSelected(false);
        }
      }
    });
    
    groupingPanel.add(mManualGroupingCheckBox, BorderLayout.NORTH);
    
    // bottom panel
    JPanel bottomPanel = new JPanel(new BorderLayout(6,6));
    groupingPanel.add(bottomPanel, BorderLayout.CENTER);
    
    // create JList on left, and button panel on right
    mImportGroupList = new ImportGroupList();
    mImportGroupList.setVisibleRowCount(5);
    mImportGroupList.setCellRenderer(new ImportGroupListRenderer());
    mImportGroupList.setBorder(BorderFactory.createEtchedBorder());
    mImportGroupList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
   
    // add a seleciton listener to the JList, so we can make sure the remove
    // button is in correct state
    mImportGroupList.addListSelectionListener(new ListSelectionListener()
    {
      public void valueChanged(ListSelectionEvent e)
      {
        refreshButtonStates();
      }
    });
        
    // create button panel
    JPanel buttonPanel = new JPanel(new GridLayout(5,1));
    mAddButton = new JButton(jEdit.getProperty("LazyImporter.LayoutOptionPane.AddImportButton"));
    mAddButton.addActionListener(this);
    mRemoveButton = new JButton(jEdit.getProperty("LazyImporter.LayoutOptionPane.RemoveButton"));
    mRemoveButton.addActionListener(this);
    mUpButton = new JButton(jEdit.getProperty("LazyImporter.LayoutOptionPane.UpButton"));
    mUpButton.addActionListener(this);
    mDownButton = new JButton(jEdit.getProperty("LazyImporter.LayoutOptionPane.DownButton"));
    mDownButton.addActionListener(this);
    mAddSpacerButton = new JButton(jEdit.getProperty("LazyImporter.LayoutOptionPane.AddBlankLineButton"));
    mAddSpacerButton.addActionListener(this);
    mAddCommentButton = new JButton("Add Comment");
    mAddCommentButton.addActionListener(this);
    buttonPanel.add(mAddButton);
    buttonPanel.add(mRemoveButton);
    buttonPanel.add(mUpButton);
    buttonPanel.add(mDownButton);
    buttonPanel.add(mAddSpacerButton);
        
    bottomPanel.add(new JScrollPane(mImportGroupList), BorderLayout.CENTER);
    bottomPanel.add(buttonPanel, BorderLayout.EAST);
  
    // set the selected index to be the first one
    mImportGroupList.setSelectedIndex(0);
    
    refreshButtonStates();
      
    return groupingPanel;
  } 
  
  private void refreshButtonStates()
  {
    mRemoveButton.setEnabled(mImportGroupList.getSelectedValue() != null &&
      mImportGroupList.getSelectedValue() != AllRemainingImportGroupElement.getInstance());
      
    mUpButton.setEnabled(mImportGroupList.getSelectedValue() != null &&
      mImportGroupList.getSelectedIndex() != 0);
    
    mDownButton.setEnabled(mImportGroupList.getSelectedValue() != null &&
      mImportGroupList.getSelectedIndex() != mImportGroupList.getModel().getSize()-1);
  }
  
  private void setProperties()
  {
    jEdit.setBooleanProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS, mGroupImportsCheckBox.isSelected());
    jEdit.setIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_LEVEL, mGroupLevelSlider.getValue());
    jEdit.setIntegerProperty(LazyImporterProperties.LAYOUT_GROUP_IMPORTS_SPACING, mGroupSpacingSlider.getValue());
    jEdit.setProperty(LazyImporterProperties.MANUAL_GROUP_IMPORTS_STRING, ImportGroupElementFactory.createElementString(((DefaultListModel)mImportGroupList.getModel()).elements()));
    jEdit.setBooleanProperty(LazyImporterProperties.MANUAL_GROUP_IMPORTS, mManualGroupingCheckBox.isSelected());
  }
  
  //
  // Inner Classes
  //
  
  /**
   * The layout previewer is a panel that contains an import "previewer" - 
   * at text area that shows some example imports.
   */
  private class LayoutPreviewer extends JPanel
  {
    private JTextArea mPreviewArea;
    private JavaClass mJavaClass;
    private int mLevel;
    private int mSpacing;
        
    public LayoutPreviewer(int level, int spacing)
    {
      mLevel = level;
      mSpacing = spacing;
      
      init();
    }
    
    public void refreshPreviewClass(int level, int spacing)
    {
      mJavaClass.setDrawingStrategy(new GroupingDrawingStrategy(level, spacing));
      mJavaClass.redrawImportStatements();
      mPreviewArea.setText(mJavaClass.getClassText());
      mPreviewArea.setCaretPosition(0);
      this.revalidate();
    }
    
    private void init()
    {
      setLayout(new BorderLayout());
      mPreviewArea = new JTextArea();
      mPreviewArea.setRows(10);
      mPreviewArea.setEditable(false);
      mPreviewArea.setLineWrap(true);
      initialiseJavaClass();
      add(new JScrollPane(mPreviewArea), BorderLayout.CENTER);
    }
        
    private void initialiseJavaClass()
    {
      mJavaClass = new JavaClass(msPreviewString);
      refreshPreviewClass(mLevel, mSpacing);
    }
      
  }

  //
  // Inner Classes
  //
  
  private class GroupLayoutChangeListener implements ChangeListener
  {
    public void stateChanged(ChangeEvent e) 
    {
      JSlider source = (JSlider)e.getSource();
      if (!source.getValueIsAdjusting()) {
        mLayoutPreviewer.refreshPreviewClass
          (mGroupLevelSlider.getValue(), mGroupSpacingSlider.getValue());
      }
    }
  }
 
 
  public class ImportGroupList extends JList
  {
    private DefaultListModel mModel;
    
    public ImportGroupList()
    {
      // create and set model
      mModel = new DefaultListModel();
      setModel(mModel);
      
      // get elements from property, and set add them to this list
      List elements = ImportGroupElementFactory.createElementsFromString
        (jEdit.getProperty(LazyImporterProperties.MANUAL_GROUP_IMPORTS_STRING, ""));
      for (Iterator i = elements.iterator(); i.hasNext();) {
        ImportGroupElement element = (ImportGroupElement) i.next();
        addElement(element);
      }
    }
  
    public void addElement(ImportGroupElement groupElement)
    {
      // either after the currently selected, or at the beginning if there is
      // none selected
      int positionToAdd = (getSelectedIndex() >= 0 ? getSelectedIndex() + 1 : 0);
      
      mModel.add(positionToAdd, groupElement);
      
      setSelectedValue(groupElement, true);
    }
    
    public void removeSelected()
    {
      if (getSelectedValue() != AllRemainingImportGroupElement.getInstance()) {
        mModel.removeElement(getSelectedValue());
      }
    }
       
    public void moveSelectedImportUp()
    {
      Object selected = getSelectedValue();
      
      // get index of this object
      int indexOfSelected = mModel.indexOf(selected);
      
      // remove and add before this index
      mModel.removeElement(selected);
      mModel.add(--indexOfSelected, selected);
    }
    
    public void moveSelectedImportDown()
    {
      Object selected = getSelectedValue();
      
      // get index of this object
      int indexOfSelected = mModel.indexOf(selected);
      
      // remove and add before this index
      mModel.removeElement(selected);
      mModel.add(++indexOfSelected, selected);
    }
  
  }
  
  private class ImportGroupListRenderer extends DefaultListCellRenderer
  {
    public Component getListCellRendererComponent
      (JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) 
    {
      Component comp = 
        super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
      
      if (comp instanceof JLabel) {
        if (value instanceof AllRemainingImportGroupElement) {
          ((JLabel)comp).setText(jEdit.getProperty("LazyImporter.LayoutOptionPane.AllRemainingImportElement.ListText"));
        } else if (value instanceof BlankLineImportGroupElement) {
          ((JLabel)comp).setText(jEdit.getProperty("LazyImporter.LayoutOptionPane.BlankLineImportElement.ListText"));
        } else if (value instanceof PrefixImportGroupElement) {
          ((JLabel)comp).setText(jEdit.getProperty("LazyImporter.LayoutOptionPane.PrefixImportElement.ListText") + value.toString());
        } else if (value instanceof CommentImportGroupElement) {
          ((JLabel)comp).setText(value.toString());
        } else {
          ((JLabel)comp).setText(comp.toString());
        }
      }
      return comp;
    }
  }
 
  //
  // ActionListener Implementation
  //
  
  public void actionPerformed(ActionEvent ae) 
  {
    if (ae.getSource() == mAddButton) {
      String importToAdd = JOptionPane.showInputDialog(this, 
        jEdit.getProperty("LazyImporter.LayoutOptionPane.AddPrefixWindow"));
      if (importToAdd != null && importToAdd.length() > 0) {
        mImportGroupList.addElement
          (ImportGroupElementFactory.createFromString(importToAdd));
      }
    } else if (ae.getSource() == mRemoveButton) {
      int selectedIndex = mImportGroupList.getSelectedIndex();
      mImportGroupList.removeSelected();
      if (mImportGroupList.getModel().getSize() > 0) {
        if (selectedIndex >= mImportGroupList.getModel().getSize()) {
          selectedIndex--;
        }
        mImportGroupList.setSelectedIndex(selectedIndex);
      }
    } else if (ae.getSource() == mUpButton) {
      Object selected = mImportGroupList.getSelectedValue();
      mImportGroupList.moveSelectedImportUp();
      mImportGroupList.setSelectedValue(selected, true);
    } else if (ae.getSource() == mDownButton) {
      Object selected = mImportGroupList.getSelectedValue();
      mImportGroupList.moveSelectedImportDown();
      mImportGroupList.setSelectedValue(selected, true);
    } else if (ae.getSource() == mAddSpacerButton) {
      mImportGroupList.addElement(new BlankLineImportGroupElement());
    } else if (ae.getSource() == mAddCommentButton) {
      String commentToAdd = JOptionPane.showInputDialog(this, "Comment:"); 
      if (commentToAdd != null && commentToAdd.length() > 0) {
        mImportGroupList.addElement
          (new CommentImportGroupElement(commentToAdd));
      }  
    }
  }
}

