/*
 * LazyImporter jEdit Plugin (JavaCommentStripper.java) 
 *  
 * Copyright (C) 2003 Matt Etheridge (matt@etheridge.org)
 * Modifications by Markus Birth (mbirth@webwriters.de)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.etheridge.lazyimporter.utility;

public class JavaCommentStripper
{
  // the buffer used to store the text of this java code while parsing
  private StringBuffer mBuffer;
  
  // line separator character
  private static String msLineSeparator = System.getProperty("line.separator");
  
  public JavaCommentStripper(String text)
  {
    mBuffer = new StringBuffer(text); 
  }
  
  public String stripComments()
  {
    // iterate through buffer, removing any comments
    int position = 0;
    boolean inStringLiteral = false;
    while (position < mBuffer.length()-1) {
      
      // check whether we are in a String
      if (mBuffer.charAt(position) == '"') {
        // check that there's no escaped quotation mark and mark string end, else mark string beginning
        if (inStringLiteral && ((position == 0) || (position>0 && mBuffer.charAt(position-1) != '\\'))) {
          inStringLiteral = false;
        } else {
          inStringLiteral = true;
        }
      }
      
      // if the previous and current characters are slashes (/), then remove
      // everything up to end of line
      if (isLineCommentStart(position) && !inStringLiteral) {
        
        int lineCommentStart = position - 1;
        // find the end of line
        while (position<mBuffer.length() && mBuffer.charAt(position) != '\n') {
          position++;
        }
        
        // once end of the line is found, remove line comment
        mBuffer.delete(lineCommentStart, position);
        position = lineCommentStart;
      }
      
      else if (isBlockCommentStart(position) && !inStringLiteral) {
        // find the end of the block comment
        int blockCommentStart = position - 1;
        while (position<mBuffer.length() && !isBlockCommentEnd(position)) {
          position++;
        }
        
        // once end of the block comment is found, remove the block comment        
        mBuffer.delete(blockCommentStart, position);
        position = blockCommentStart;
      } else {
        position++;
      }
    }
    return mBuffer.toString();
  }
  
  //
  // Private Helper Methods
  //
  
  private boolean isLineCommentStart(int position)
  {
    // if we are at the start of the buffer, then this obviously cannot
    // be the start of a line comment (as it takes two characters)
    if (position == 0) {
      return false;
    }
    
    // if the previous and current characters are slashes (/) then this is
    // the start of a line comment
    if (mBuffer.charAt(position) == '/' && 
        mBuffer.charAt(position - 1) == '/') {
      return true;
    }
    
    return false;
  }
  
  private boolean isBlockCommentStart(int position)
  {
    // if we are at the start of the buffer, then this obviously cannot
    // be the start of a line comment (as it takes two characters)
    if (position == 0) {
      return false;
    }
    
    // if the current character is a * and the previous character is a / then
    // this is the start of a block comment
    if (mBuffer.charAt(position) == '*' && 
        mBuffer.charAt(position - 1) == '/') {
      return true;
    }
    
    return false;
  }
  
  private boolean isBlockCommentEnd(int position)
  {
    // if we are at the start of the buffer this can't be the end
    if (position == 0) {
      return false;
    }
    
    // if the current character is a * and the previous character is a / then
    // this is the start of a block comment
    if (mBuffer.charAt(position) == '/' && 
        mBuffer.charAt(position - 1) == '*') {
      return true;
    }
    
    return false;
  }
 
}
